
import os
from PyQt4.QtCore import *
from PyQt4.QtGui import *

from sfa.util.rspecHelper import RSpec
from sface.sfahelper import *
from sface.config import config
from sface.sfiprocess import SfiProcess
from sface.screens.sfascreen import SfaScreen

already_in_nodes = []

node_status = { "in": "Already Selected",
                "out": "Not Selected",
                "add": "To be Added",
                "remove": "To be Removed"}

class NodeView(QTreeView):
    def __init__(self, parent):
        QTreeView.__init__(self, parent)

        self.setAnimated(True)
        self.setItemsExpandable(True)
        self.setRootIsDecorated(True)
        self.setAlternatingRowColors(True)
#        self.setSelectionMode(self.MultiSelection)
        self.setAttribute(Qt.WA_MacShowFocusRect, 0)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setToolTip("Double click on a row to change its status")

    def mouseDoubleClickEvent(self, event):
        index = self.currentIndex()
        model = index.model()
        status_index = model.index(index.row(), 2, index.parent())
        status_data = status_index.data().toString()
        hostname_index = model.index(index.row(), 1, index.parent())
        hostname_data = hostname_index.data().toString()

        if status_data == node_status['in']:
            model.setData(status_index, QString(node_status['remove']))
        elif status_data == node_status['out']:
            model.setData(status_index, QString(node_status['add']))
        elif status_data in (node_status['add'], node_status['remove']):
            if hostname_data in already_in_nodes: model.setData(status_index, QString(node_status['in']))
            else: model.setData(status_index, QString(node_status['out']))

        model.emit(SIGNAL("dataChanged(QModelIndex, QModelIndex)"), hostname_index, hostname_index)

    def currentChanged(self, current, previous):
        model = current.model()
        hostname_index = model.index(current.row(), 1, current.parent())
        hostname_data = hostname_index.data().toString()
        self.emit(SIGNAL('hostnameClicked(QString)'), hostname_data)
        
                

class NodeNameDelegate(QStyledItemDelegate):
    def __init__(self, parent):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        model = index.model()
        data = "%s" % index.data().toString()
        status_index = model.index(index.row(), 2, index.parent())
        status_data = status_index.data().toString()

        if status_data not in (node_status['in'], node_status['remove'], node_status['add']):
            # default view
            QStyledItemDelegate.paint(self, painter, option, index)
            return

        fm = QFontMetrics(option.font)
        rect = option.rect
        rect.setWidth(fm.width(QString(data)) + 8)
        rect.setHeight(rect.height() - 2)
        rect.setX(rect.x() + 4)
        x, y, h, w = rect.x(), rect.y(), rect.height(), rect.width() 

        path = QPainterPath()
        path.addRoundedRect(x, y, w, h, 4, 4)

        painter.save()
        painter.setRenderHint(QPainter.Antialiasing)
        painter.drawRoundedRect(rect, 4, 4)

        if status_data == node_status['in']: # already in the slice
            painter.fillPath(path, QColor.fromRgb(0, 250, 0))
            painter.setPen(QColor.fromRgb(0, 0, 0))
            painter.drawText(option.rect, 0, QString(data))

        elif status_data == node_status['add']: # newly added to the slice
            painter.fillPath(path, QColor.fromRgb(0, 250, 0))
            painter.setPen(QColor.fromRgb(0, 0, 0))
            painter.drawText(option.rect, 0, QString(data))
            painter.drawRect(x + w + 10, y + 3, 10, 10)
            painter.fillRect(x + w + 10, y + 3, 10, 10, QColor.fromRgb(0, 250, 0))

        elif status_data == node_status['remove']: # removed from the slice
            painter.fillPath(path, QColor.fromRgb(250, 0, 0))
            painter.setPen(QColor.fromRgb(0, 0, 0))
            painter.drawText(option.rect, 0, QString(data))
            painter.drawRect(x + w + 10, y + 3, 10, 10)
            painter.fillRect(x + w + 10, y + 3, 10, 10, QColor.fromRgb(250, 0, 0))

        painter.restore()


class TreeItem:
    def __init__(self, data, parent=None):
        self.parentItem = parent
        self.itemData = data
        self.childItems = []

    def clear(self):
        for child in self.childItems:
            child.clear()
            del child
        del self.childItems
        self.childItems = []

    def allChildItems(self):
        all = []
        for c in self.childItems:
            all.append(c)
            if c.childItems:
                for cc in c.childItems:
                    all.append(cc)
        return all

    def appendChild(self, child):
        self.childItems.append(child)

    def child(self, row):
        return self.childItems[row]
    
    def childCount(self):
        return len(self.childItems)

    def childNumber(self):
        if self.parentItem:
            return self.parentItem.childItems.index(self)
        return 0

    def columnCount(self):
        return len(self.itemData)

    def data(self, column):
        return self.itemData[column]

    def insertChildren(self, position, count, columns):
        if position < 0 or position > len(self.childItems):
            return False
        
        for row in range(count):
            data = self.data(columns)
            item = TreeItem(data, self)
            self.childItems.insert(position, item)

        return True

    def insertColumns(self, position, columns):
        if position < 0 or position > len(self.itemData):
            return False

        for column in range(columns):
            self.itemData.insert(position, QVariant())
        
        for child in self.childItems:
            child.insertColumns(position, columns)
        
        return True

    def setData(self, column, value):
        if column < 0 or column >= len(self.itemData):
            return False

        self.itemData[column] = value
        return True
    
    def parent(self):
        return self.parentItem


class NodeModel(QAbstractItemModel):
    def __init__(self, parent):
        QAbstractItemModel.__init__(self, parent)
        self.__initRoot()

    def clear(self):
        self.rootItem.clear()
        self.__initRoot()

    def __initRoot(self):
        self.rootItem = TreeItem([QString("Testbed"), QString("Hostname"), QString("Status")])

    def getItem(self, index):
        if index.isValid():
            item = index.internalPointer()
            if isinstance(item, TreeItem):
                return item
        return self.rootItem

    def headerData(self, section, orientation, role):
        if orientation == Qt.Horizontal and role in (Qt.DisplayRole, Qt.EditRole):
            return self.rootItem.data(section)
        return QVariant()

    def index(self, row, column, parent):
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        parentItem = self.getItem(parent)
        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QModelIndex()

    def insertColumns(self, position, columns, parent):
        self.beginInsertColumns(parent, position, position + columns -1)
        ret = self.rootItem.insertColumns(position, columns)
        self.endInsertColumns()
        return ret

    def insertRows(self, position, rows, parent):
        parentItem = self.getItem(parent)
        self.beginInsertRows(parent, position, position + rows -1)
        ret = parentItem.insertChildren(position, rows, self.rootItem.columnCount())
        self.endInsertRows()
        return ret

    def parent(self, index):
        if not index.isValid():
            return QModelIndex()

        childItem = self.getItem(index)
        parentItem = childItem.parent()
        if parentItem is self.rootItem:
            return QModelIndex()

        return self.createIndex(parentItem.childNumber(), 0, parentItem)

    def rowCount(self, parent=QModelIndex()):
        parentItem = self.getItem(parent)
        return parentItem.childCount()

    def columnCount(self, parent=None):
        return self.rootItem.columnCount()

    def data(self, index, role):
        if not index.isValid():
            return QVariant()

        if role != Qt.DisplayRole and role != Qt.EditRole:
            return QVariant()

        item = self.getItem(index)
        return item.data(index.column())

    def flags(self, index):
        if not index.isValid():
            return 0
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable# | Qt.ItemIsEditable

    def setData(self, index, value, role):
        if role != Qt.EditRole:
            return False

        item = self.getItem(index)
        ret = item.setData(index.column(), value)
        if ret:
            self.emit(SIGNAL("dataChanged(QModelIndex, QModelIndex)"), index, index)
        return ret


class SliceWidget(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self, parent)

        self.network_names = []
        self.process = SfiProcess(self)

        self.slicename = QLabel("", self)
        self.updateSliceName()
        self.slicename.setScaledContents(False)
        searchlabel = QLabel ("Search: ", self)
        searchlabel.setScaledContents(False)
        searchbox = QLineEdit(self)
        searchbox.setAttribute(Qt.WA_MacShowFocusRect, 0)

        toplayout = QHBoxLayout()
        toplayout.addWidget(self.slicename, 0, Qt.AlignLeft)
        toplayout.addStretch()
        toplayout.addWidget(searchlabel, 0, Qt.AlignRight)
        toplayout.addWidget(searchbox, 0, Qt.AlignRight)

        self.nodeView = NodeView(self)
        self.nodeModel = NodeModel(self)
        self.filterModel = QSortFilterProxyModel(self) # enable filtering

        self.nodeNameDelegate = NodeNameDelegate(self)

        refresh = QPushButton("Update Slice Data", self)
        refresh.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        submit = QPushButton("Submit", self)
        submit.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)

        bottomlayout = QHBoxLayout()
        bottomlayout.addWidget(refresh, 0, Qt.AlignLeft)
        bottomlayout.addStretch()
        bottomlayout.addWidget(submit, 0, Qt.AlignRight)

        layout = QVBoxLayout()
        layout.addLayout(toplayout)
        layout.addWidget(self.nodeView)
        layout.addLayout(bottomlayout)
        self.setLayout(layout)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.connect(refresh, SIGNAL('clicked()'), self.refresh)
        self.connect(submit, SIGNAL('clicked()'), self.submit)
        self.connect(searchbox, SIGNAL('textChanged(QString)'), self.filter)
        self.connect(self.nodeView, SIGNAL('hostnameClicked(QString)'),
                     self.nodeSelectionChanged)

        self.updateView()

    def submitFinished(self):
        self.setStatus("<font color='green'>Slice data submitted.</font>")
        QTimer.singleShot(1000, self.refresh)

    def refreshFinished(self):
        self.setStatus("<font color='green'>Slice data updated.</font>", timeout=5000)
        self.updateView()
        self.parent().signalAll("rspecUpdated")

    def readSliceRSpec(self):
        rspec_file = config.getSliceRSpecFile()
        if os.path.exists(rspec_file):
            xml = open(rspec_file).read()
            return xml
        return None

    def setStatus(self, msg, timeout=None):
        self.parent().setStatus(msg, timeout)

    def checkRunningProcess(self):
        if self.process.isRunning():
            self.setStatus("<font color='red'>There is already a process running. Please wait.</font>")
            return True
        return False

    def filter(self, filter_string):
        # for hierarchical models QSortFilterProxyModel applies the
        # sort recursively. if the parent doesn't match the criteria
        # we won't be able to match the children. so we need to match
        # parent (by matching the network_names)
        networks = ["^%s$" % n for n in self.network_names]
        filters = networks + [str(filter_string)]
        self.filterModel.setFilterRegExp(QRegExp('|'.join(filters)))

    def submit(self):
        if self.checkRunningProcess():
            return

        rspec = RSpec(self.readSliceRSpec())
        
        no_change = True
        all_child = self.nodeModel.rootItem.allChildItems()
        for c in all_child:
            testbed, hostname, status = c.itemData
            if isinstance(status, QVariant):
                status = status.toString()

            if status == node_status['add']:
                rspec.add_sliver(hostname)
                no_change = False
            elif str(status) == node_status['remove']:
                rspec.remove_sliver(hostname)
                no_change = False

        if no_change:
            self.setStatus("<font color=red>No change in slice data. Not submitting!</font>", timeout=3000)
            return

        self.disconnect(self.process, SIGNAL('finished()'), self.refreshFinished)
        self.connect(self.process, SIGNAL('finished()'), self.submitFinished)

        self.process.applyRSpec(rspec)
        self.setStatus("Sending slice data (RSpec). This will take some time...")
        

    def refresh(self):
        if not config.getSlice():
            self.setStatus("<font color='red'>Slice not set yet!</font>")
            return

        if self.process.isRunning():
            self.setStatus("<font color='red'>There is already a process running. Please wait.</font>")
            return

        self.disconnect(self.process, SIGNAL('finished()'), self.submitFinished)
        self.connect(self.process, SIGNAL('finished()'), self.refreshFinished)

        self.process.getRSpecFromSM()
        self.setStatus("Updating slice data. This will take some time...")

    def updateView(self):
        global already_in_nodes
        already_in_nodes = []
        self.network_names = []
        self.nodeModel.clear()
        
        rspec_string = self.readSliceRSpec()
        if not rspec_string:
            return None

        networks = rspec_get_networks(rspec_string)
        for network in networks:
            self.network_names.append(network)
            networkItem = TreeItem([QString(network), QString(""), QString("")], self.nodeModel.rootItem)

            all_nodes = rspec_get_nodes_from_network(rspec_string, network)
            sliver_nodes = rspec_get_sliver_nodes_from_network(rspec_string, network)
            available_nodes = filter(lambda x:x not in sliver_nodes, all_nodes)

            already_in_nodes += sliver_nodes

            for node in sliver_nodes:
                nodeItem = TreeItem([QString(""), QString("%s" % node), QString(node_status['in'])], networkItem)
                networkItem.appendChild(nodeItem)

            for node in available_nodes:
                nodeItem = TreeItem([QString(""), QString(node), QString(node_status['out'])], networkItem)
                networkItem.appendChild(nodeItem)

            self.nodeModel.rootItem.appendChild(networkItem)

        self.filterModel.setSourceModel(self.nodeModel)
        self.filterModel.setFilterKeyColumn(-1)
        self.filterModel.setDynamicSortFilter(True)

        self.nodeView.setItemDelegateForColumn(1, self.nodeNameDelegate)
        self.nodeView.setModel(self.filterModel)
        self.nodeView.expandAll()
        self.nodeView.resizeColumnToContents(1)

    def updateSliceName(self):
        self.slicename.setText("Slice : %s" % (config.getSlice() or "None"))

    def nodeSelectionChanged(self, hostname):
        self.parent().nodeSelectionChanged(hostname)

class MainScreen(SfaScreen):
    def __init__(self, parent):
        SfaScreen.__init__(self, parent)

        slice = SliceWidget(self)
        self.init(slice, "Main Window", "OneLab SFA crawler")

    def rspecUpdated(self):
        self.mainwin.rspecWindow.updateView()
        
    def configurationChanged(self):
        self.widget.updateSliceName()
        self.widget.updateView()
        self.mainwin.rspecWindow.updateView()

    def nodeSelectionChanged(self, hostname):
        self.mainwin.nodeSelectionChanged(hostname)
